from __future__ import annotations

from typing import TYPE_CHECKING, Any, cast

from litestar.constants import SKIP_VALIDATION_NAMES
from litestar.exceptions import ImproperlyConfiguredException
from litestar.params import DependencyKwarg
from litestar.types import Empty

if TYPE_CHECKING:
    from litestar.utils.signature import ParsedSignature

    from .model import SignatureModel

__all__ = ("create_type_overrides", "validate_signature_dependencies", "get_signature_model")


def get_signature_model(value: Any) -> type[SignatureModel]:
    """Retrieve and validate the signature model from a provider or handler."""
    try:
        return cast("type[SignatureModel]", value.signature_model)
    except AttributeError as e:  # pragma: no cover
        raise ImproperlyConfiguredException(f"The 'signature_model' attribute for {value} is not set") from e


def create_type_overrides(parsed_signature: ParsedSignature, has_data_dto: bool) -> dict[str, Any]:
    """Create typing overrides for field definitions.

    Args:
        parsed_signature: A parsed function signature.
        has_data_dto: Whether the signature contains a data DTO.

    Returns:
        A dictionary of typing overrides
    """
    type_overrides = {}
    for field_definition in parsed_signature.parameters.values():
        if field_definition.name in SKIP_VALIDATION_NAMES or (
            isinstance(field_definition.kwarg_definition, DependencyKwarg)
            and field_definition.kwarg_definition.skip_validation
        ):
            type_overrides[field_definition.name] = Any

        if has_data_dto and "data" in parsed_signature.parameters:
            type_overrides["data"] = Any

    return type_overrides


def validate_signature_dependencies(
    dependency_name_set: set[str], fn_name: str, parsed_signature: ParsedSignature
) -> set[str]:
    """Validate dependencies of ``parsed_signature``.

    Args:
        dependency_name_set: A set of dependency names
        fn_name: A callable's name.
        parsed_signature: A parsed signature.

    Returns:
        A set of validated dependency names.
    """
    dependency_names: set[str] = set(dependency_name_set)

    for parameter in parsed_signature.parameters.values():
        if isinstance(parameter.kwarg_definition, DependencyKwarg) and parameter.name not in dependency_name_set:
            if not parameter.is_optional and parameter.default is Empty:
                raise ImproperlyConfiguredException(
                    f"Explicit dependency '{parameter.name}' for '{fn_name}' has no default value, "
                    f"or provided dependency."
                )
            dependency_names.add(parameter.name)
    return dependency_names



from litestar.utils import warn_deprecation


def __getattr__(attr_name: str) -> object:
    if "MsgspecDTO" in attr_name:
        from litestar.dto.msgspec_dto_factory import MsgspecDTO

        warn_deprecation(
            deprecated_name="litestar.contrib.msgspec.MsgspecDTO",
            version="2.0b3",
            kind="import",
            removal_in="2.0",
            info="importing 'MsgspecDTO' from 'litestar.contrib.msgspec' is deprecated, please"
            "import it from 'litestar.dto.factory' instead",
        )

        globals()[attr_name] = MsgspecDTO
        return MsgspecDTO
    raise AttributeError(f"module {__name__!r} has no attribute {attr_name!r}")



"""DTO backends do the heavy lifting of decoding and validating raw bytes into domain models, and
back again, to bytes.
"""
from __future__ import annotations

import secrets
from typing import TYPE_CHECKING, Any, Final, cast

from msgspec import Struct, convert

from litestar.dto._types import (
    CollectionType,
    CompositeType,
    MappingType,
    NestedFieldInfo,
    SimpleType,
    TransferDTOFieldDefinition,
    TupleType,
    UnionType,
)
from litestar.dto._utils import (
    RenameStrategies,
    build_annotation_for_backend,
    create_struct_for_field_definitions,
    should_exclude_field,
    should_ignore_field,
    should_mark_private,
    transfer_data,
)
from litestar.dto.data_structures import DTOData
from litestar.dto.field import Mark
from litestar.enums import RequestEncodingType
from litestar.exceptions import SerializationException
from litestar.serialization import decode_json, decode_msgpack
from litestar.typing import FieldDefinition
from litestar.utils.helpers import get_fully_qualified_class_name

if TYPE_CHECKING:
    from typing import AbstractSet, Callable, Collection, Generator

    from litestar._openapi.schema_generation import SchemaCreator
    from litestar.dto import DTOConfig
    from litestar.dto._types import FieldDefinitionsType
    from litestar.dto.data_structures import DTOFieldDefinition
    from litestar.dto.interface import ConnectionContext
    from litestar.dto.types import ForType
    from litestar.openapi.spec import Reference, Schema
    from litestar.types.serialization import LitestarEncodableType

__all__ = ("DTOBackend", "BackendContext")


class BackendContext:
    """Context required by DTO backends to perform their work."""

    __slots__ = (
        "config",
        "dto_for",
        "field_definition_generator",
        "is_nested_field_predicate",
        "model_type",
        "field_definition",
        "wrapper_attribute_name",
    )

    def __init__(
        self,
        dto_config: DTOConfig,
        dto_for: ForType,
        field_definition: FieldDefinition,
        field_definition_generator: Callable[[Any], Generator[DTOFieldDefinition, None, None]],
        is_nested_field_predicate: Callable[[FieldDefinition], bool],
        model_type: type[Any],
        wrapper_attribute_name: str | None,
    ) -> None:
        """Create a backend context.

        Args:
            dto_config: DTO config.
            dto_for: "data" or "return"
            field_definition: Parsed type.
            field_definition_generator: Generator that produces
                :class:`FieldDefinition <.dto.factory.types.FieldDefinition>` instances given ``model_type``.
            is_nested_field_predicate: Function that detects if a field is nested.
            model_type: Model type.
            wrapper_attribute_name: If the data that DTO should operate upon is wrapped in a generic datastructure, this is the
                name of the attribute that the data is stored in.
        """
        self.config: Final[DTOConfig] = dto_config
        self.dto_for: Final[ForType] = dto_for
        self.field_definition: Final[FieldDefinition] = field_definition
        self.field_definition_generator: Final[
            Callable[[Any], Generator[DTOFieldDefinition, None, None]]
        ] = field_definition_generator
        self.is_nested_field_predicate: Final[Callable[[FieldDefinition], bool]] = is_nested_field_predicate
        self.model_type: Final[type[Any]] = model_type
        self.wrapper_attribute_name = wrapper_attribute_name


class NestedDepthExceededError(Exception):
    """Raised when a nested type exceeds the maximum allowed depth.

    Not an exception that is intended to be raised into userland, rather a signal to the process that is iterating over
    the field definitions that the current field should be skipped.
    """


class DTOBackend:
    __slots__ = (
        "annotation",
        "context",
        "dto_data_type",
        "parsed_field_definitions",
        "reverse_name_map",
        "transfer_model_type",
    )

    def __init__(self, context: BackendContext) -> None:
        """Create dto backend instance.

        Args:
            context: context of the type represented by this backend.
        """
        self.context = context
        self.parsed_field_definitions = self.parse_model(
            model_type=context.model_type, exclude=context.config.exclude, include=context.config.include
        )
        self.transfer_model_type = self.create_transfer_model_type(
            get_fully_qualified_class_name(context.model_type), self.parsed_field_definitions
        )
        self.dto_data_type: type[DTOData] | None = None
        if context.field_definition.is_subclass_of(DTOData):
            self.dto_data_type = context.field_definition.annotation
            annotation = self.context.field_definition.inner_types[0].annotation
        else:
            annotation = context.field_definition.annotation
        self.annotation = build_annotation_for_backend(annotation, self.transfer_model_type)

    def parse_model(
        self, model_type: Any, exclude: AbstractSet[str], include: AbstractSet[str], nested_depth: int = 0
    ) -> FieldDefinitionsType:
        """Reduce :attr:`model_type` to :class:`FieldDefinitionsType`.

        .. important::
            Implementations must respect the :attr:`config` object. For example:
                - fields marked private must never be included in the field definitions.
                - if a ``purpose`` is declared, then read-only fields must be taken into account.
                - field renaming must be implemented.
                - additional fields must be included, subject to ``purpose``.
                - nested depth and nested recursion depth must be adhered to.

        Returns:
            Fields for data transfer.

            Key is the name of the new field, and value is a tuple of type and default value pairs.

            Add a new field called "new_field", that is a string, and required:
            {"new_field": (str, ...)}

            Add a new field called "new_field", that is a string, and not-required:
            {"new_field": (str, "default")}

            Add a new field called "new_field", that may be `None`:
            {"new_field": (str | None, None)}
        """
        defined_fields = []
        for field_definition in self.context.field_definition_generator(model_type):
            if should_ignore_field(field_definition, self.context.dto_for):
                continue

            if should_mark_private(field_definition, self.context.config.underscore_fields_private):
                field_definition.dto_field.mark = Mark.PRIVATE

            try:
                transfer_type = self._create_transfer_type(
                    field_definition=field_definition,
                    exclude=exclude,
                    include=include,
                    field_name=field_definition.name,
                    unique_name=field_definition.unique_name(),
                    nested_depth=nested_depth,
                )
            except NestedDepthExceededError:
                continue

            if rename := self.context.config.rename_fields.get(field_definition.name):
                serialization_name = rename
            elif self.context.config.rename_strategy:
                serialization_name = RenameStrategies(self.context.config.rename_strategy)(field_definition.name)
            else:
                serialization_name = field_definition.name

            transfer_field_definition = TransferDTOFieldDefinition.from_dto_field_definition(
                field_definition=field_definition,
                serialization_name=serialization_name,
                transfer_type=transfer_type,
                is_partial=self.context.config.partial,
                is_excluded=should_exclude_field(
                    field_definition=field_definition, exclude=exclude, include=include, dto_for=self.context.dto_for
                ),
            )
            defined_fields.append(transfer_field_definition)
        return tuple(defined_fields)

    def create_transfer_model_type(self, unique_name: str, field_definitions: FieldDefinitionsType) -> type[Struct]:
        """Create a model for data transfer.

        Args:
            unique_name: name for the type that should be unique across all transfer types.
            field_definitions: field definitions for the container type.

        Returns:
            A ``BackendT`` class.
        """
        fqn_uid: str = self._gen_unique_name_id(unique_name)
        struct = create_struct_for_field_definitions(fqn_uid, field_definitions)
        setattr(struct, "__schema_name__", unique_name)
        return struct

    def parse_raw(self, raw: bytes, connection_context: ConnectionContext) -> Struct | Collection[Struct]:
        """Parse raw bytes into transfer model type.

        Args:
            raw: bytes
            connection_context: Information about the active connection.

        Returns:
            The raw bytes parsed into transfer model type.
        """

        if connection_context.request_encoding_type not in [RequestEncodingType.JSON, RequestEncodingType.MESSAGEPACK]:
            raise SerializationException(
                f"Unsupported request encoding type: '{connection_context.request_encoding_type}'"
            )

        if connection_context.request_encoding_type == RequestEncodingType.JSON:
            result = decode_json(value=raw, target_type=self.annotation, type_decoders=connection_context.type_decoders)
        else:
            result = decode_msgpack(
                value=raw, target_type=self.annotation, type_decoders=connection_context.type_decoders
            )

        return cast("Struct | Collection[Struct]", result)

    def parse_builtins(self, builtins: Any, connection_context: ConnectionContext) -> Any:
        """Parse builtin types into transfer model type.

        Args:
            builtins: Builtin type.
            connection_context: Information about the active connection.

        Returns:
            The builtin type parsed into transfer model type.
        """
        return convert(
            obj=builtins, type=self.annotation, dec_hook=connection_context.default_deserializer, strict=False
        )

    def populate_data_from_builtins(self, builtins: Any, connection_context: ConnectionContext) -> Any:
        """Populate model instance from builtin types.

        Args:
            builtins: Builtin type.
            connection_context: Information about the active connection.

        Returns:
            Instance or collection of ``model_type`` instances.
        """
        if self.dto_data_type:
            return self.dto_data_type(
                backend=self,
                data_as_builtins=transfer_data(
                    destination_type=dict,
                    source_data=self.parse_builtins(builtins, connection_context),
                    field_definitions=self.parsed_field_definitions,
                    dto_for="data",
                    field_definition=self.context.field_definition,
                ),
            )
        return self.transfer_data_from_builtins(self.parse_builtins(builtins, connection_context))

    def transfer_data_from_builtins(self, builtins: Any) -> Any:
        """Populate model instance from builtin types.

        Args:
            builtins: Builtin type.

        Returns:
            Instance or collection of ``model_type`` instances.
        """
        return transfer_data(
            self.context.model_type, builtins, self.parsed_field_definitions, "data", self.context.field_definition
        )

    def populate_data_from_raw(self, raw: bytes, connection_context: ConnectionContext) -> Any:
        """Parse raw bytes into instance of `model_type`.

        Args:
            raw: bytes
            connection_context: Information about the active connection.

        Returns:
            Instance or collection of ``model_type`` instances.
        """
        if self.dto_data_type:
            return self.dto_data_type(
                backend=self,
                data_as_builtins=transfer_data(
                    dict,
                    self.parse_raw(raw, connection_context),
                    self.parsed_field_definitions,
                    "data",
                    self.context.field_definition,
                ),
            )
        return transfer_data(
            self.context.model_type,
            self.parse_raw(raw, connection_context),
            self.parsed_field_definitions,
            "data",
            self.context.field_definition,
        )

    def encode_data(self, data: Any, connection_context: ConnectionContext) -> LitestarEncodableType:
        """Encode data into a ``LitestarEncodableType``.

        Args:
            data: Data to encode.
            connection_context: Information about the active connection.

        Returns:
            Encoded data.
        """
        if self.context.wrapper_attribute_name:
            setattr(
                data,
                self.context.wrapper_attribute_name,
                transfer_data(
                    destination_type=self.transfer_model_type,
                    source_data=getattr(data, self.context.wrapper_attribute_name),
                    field_definitions=self.parsed_field_definitions,
                    dto_for="return",
                    field_definition=self.context.field_definition,
                ),
            )
            # cast() here because we take for granted that whatever ``data`` is, it must be something
            # that litestar can natively encode.
            return cast("LitestarEncodableType", data)

        return transfer_data(
            destination_type=self.transfer_model_type,
            source_data=data,
            field_definitions=self.parsed_field_definitions,
            dto_for="return",
            field_definition=self.context.field_definition,
        )

    def create_openapi_schema(self, schema_creator: SchemaCreator) -> Reference | Schema:
        """Create an openAPI schema for the given DTO."""
        return schema_creator.for_field_definition(
            FieldDefinition.from_annotation(self.annotation), dto_for=self.context.dto_for
        )

    def _create_transfer_type(
        self,
        field_definition: FieldDefinition,
        exclude: AbstractSet[str],
        include: AbstractSet[str],
        field_name: str,
        unique_name: str,
        nested_depth: int,
    ) -> CompositeType | SimpleType:
        exclude = _filter_nested_field(exclude, field_name)
        include = _filter_nested_field(include, field_name)

        if field_definition.is_union:
            return self._create_union_type(
                field_definition=field_definition,
                exclude=exclude,
                include=include,
                unique_name=unique_name,
                nested_depth=nested_depth,
            )

        if field_definition.is_tuple:
            if len(field_definition.inner_types) == 2 and field_definition.inner_types[1].annotation is Ellipsis:
                return self._create_collection_type(
                    field_definition=field_definition,
                    exclude=exclude,
                    include=include,
                    unique_name=unique_name,
                    nested_depth=nested_depth,
                )
            return self._create_tuple_type(
                field_definition=field_definition,
                exclude=exclude,
                include=include,
                unique_name=unique_name,
                nested_depth=nested_depth,
            )

        if field_definition.is_mapping:
            return self._create_mapping_type(
                field_definition=field_definition,
                exclude=exclude,
                include=include,
                unique_name=unique_name,
                nested_depth=nested_depth,
            )

        if field_definition.is_non_string_collection:
            return self._create_collection_type(
                field_definition=field_definition,
                exclude=exclude,
                include=include,
                unique_name=unique_name,
                nested_depth=nested_depth,
            )

        transfer_model: NestedFieldInfo | None = None
        if self.context.is_nested_field_predicate(field_definition):
            if nested_depth == self.context.config.max_nested_depth:
                raise NestedDepthExceededError()

            nested_field_definitions = self.parse_model(
                model_type=field_definition.annotation, exclude=exclude, include=include, nested_depth=nested_depth + 1
            )
            transfer_model = NestedFieldInfo(
                model=self.create_transfer_model_type(unique_name, nested_field_definitions),
                field_definitions=nested_field_definitions,
            )

        return SimpleType(field_definition, nested_field_info=transfer_model)

    def _create_collection_type(
        self,
        field_definition: FieldDefinition,
        exclude: AbstractSet[str],
        include: AbstractSet[str],
        unique_name: str,
        nested_depth: int,
    ) -> CollectionType:
        inner_types = field_definition.inner_types
        inner_type = self._create_transfer_type(
            field_definition=inner_types[0] if inner_types else FieldDefinition.from_annotation(Any),
            exclude=exclude,
            include=include,
            field_name="0",
            unique_name=_enumerate_name(unique_name, 0),
            nested_depth=nested_depth,
        )
        return CollectionType(
            field_definition=field_definition, inner_type=inner_type, has_nested=_determine_has_nested(inner_type)
        )

    def _create_mapping_type(
        self,
        field_definition: FieldDefinition,
        exclude: AbstractSet[str],
        include: AbstractSet[str],
        unique_name: str,
        nested_depth: int,
    ) -> MappingType:
        inner_types = field_definition.inner_types
        key_type = self._create_transfer_type(
            field_definition=inner_types[0] if inner_types else FieldDefinition.from_annotation(Any),
            exclude=exclude,
            include=include,
            field_name="0",
            unique_name=_enumerate_name(unique_name, 0),
            nested_depth=nested_depth,
        )
        value_type = self._create_transfer_type(
            field_definition=inner_types[1] if inner_types else FieldDefinition.from_annotation(Any),
            exclude=exclude,
            include=include,
            field_name="1",
            unique_name=_enumerate_name(unique_name, 1),
            nested_depth=nested_depth,
        )
        return MappingType(
            field_definition=field_definition,
            key_type=key_type,
            value_type=value_type,
            has_nested=_determine_has_nested(key_type) or _determine_has_nested(value_type),
        )

    def _create_tuple_type(
        self,
        field_definition: FieldDefinition,
        exclude: AbstractSet[str],
        include: AbstractSet[str],
        unique_name: str,
        nested_depth: int,
    ) -> TupleType:
        inner_types = tuple(
            self._create_transfer_type(
                field_definition=inner_type,
                exclude=exclude,
                include=include,
                field_name=str(i),
                unique_name=_enumerate_name(unique_name, i),
                nested_depth=nested_depth,
            )
            for i, inner_type in enumerate(field_definition.inner_types)
        )
        return TupleType(
            field_definition=field_definition,
            inner_types=inner_types,
            has_nested=any(_determine_has_nested(t) for t in inner_types),
        )

    def _create_union_type(
        self,
        field_definition: FieldDefinition,
        exclude: AbstractSet[str],
        include: AbstractSet[str],
        unique_name: str,
        nested_depth: int,
    ) -> UnionType:
        inner_types = tuple(
            self._create_transfer_type(
                field_definition=inner_type,
                exclude=exclude,
                include=include,
                field_name=str(i),
                unique_name=_enumerate_name(unique_name, i),
                nested_depth=nested_depth,
            )
            for i, inner_type in enumerate(field_definition.inner_types)
        )
        return UnionType(
            field_definition=field_definition,
            inner_types=inner_types,
            has_nested=any(_determine_has_nested(t) for t in inner_types),
        )

    @staticmethod
    def _gen_unique_name_id(unique_name: str) -> str:
        # Generate a unique ID
        # Convert the ID to a short alphanumeric string
        return f"{unique_name}-{secrets.token_hex(8)}"


def _filter_nested_field(field_name_set: AbstractSet[str], field_name: str) -> AbstractSet[str]:
    """Filter a nested field name."""
    return {split[1] for s in field_name_set if (split := s.split(".", 1))[0] == field_name and len(split) > 1}


def _enumerate_name(name: str, index: int) -> str:
    """Enumerate ``name`` with ``index``."""
    return f"{name}_{index}"


def _determine_has_nested(transfer_type: SimpleType | CompositeType) -> bool:
    """Determine if a transfer type has nested types."""
    if isinstance(transfer_type, SimpleType):
        return bool(transfer_type.nested_field_info)
    return transfer_type.has_nested



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Callable, Literal

    from typing_extensions import TypeAlias

__all__ = ("ForType", "RenameStrategy")

ForType: TypeAlias = "Literal['data', 'return']"
"""Type for parameters that express the purpose of DTO application."""
RenameStrategy: TypeAlias = 'Literal["lower", "upper", "camel", "pascal"] | Callable[[str], str]'
"""A pre-defined strategy or a custom callback for converting DTO field names."""



from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING, Generic, TypeVar

from litestar.dto._backend import BackendContext, DTOBackend
from litestar.dto._utils import (
    get_dto_config_from_annotated_type,
    resolve_generic_wrapper_type,
    resolve_model_type,
)
from litestar.dto.config import DTOConfig
from litestar.dto.interface import ConnectionContext, DTOInterface
from litestar.exceptions.dto_exceptions import InvalidAnnotationException
from litestar.typing import FieldDefinition

if TYPE_CHECKING:
    from typing import Any, ClassVar, Collection, Generator

    from typing_extensions import Self

    from litestar._openapi.schema_generation import SchemaCreator
    from litestar.dto.data_structures import DTOFieldDefinition
    from litestar.dto.interface import HandlerContext
    from litestar.dto.types import ForType
    from litestar.enums import RequestEncodingType
    from litestar.openapi.spec import Reference, Schema
    from litestar.types.serialization import LitestarEncodableType

__all__ = ("AbstractDTOFactory",)

T = TypeVar("T")


class AbstractDTOFactory(DTOInterface, Generic[T]):
    """Base class for DTO types."""

    __slots__ = ("connection_context",)

    config: ClassVar[DTOConfig]
    """Config objects to define properties of the DTO."""
    model_type: ClassVar[type[Any]]
    """If ``annotation`` is an iterable, this is the inner type, otherwise will be the same as ``annotation``."""

    _type_backend_map: ClassVar[dict[tuple[ForType, FieldDefinition, RequestEncodingType | str | None], DTOBackend]]
    _handler_backend_map: ClassVar[dict[tuple[ForType, str], DTOBackend]]

    def __init__(self, connection_context: ConnectionContext) -> None:
        """Create an AbstractDTOFactory type.

        Args:
            connection_context: A :class:`ConnectionContext <.ConnectionContext>` instance, which provides
                information about the connection.
        """
        super().__init__(connection_context=connection_context)

    def __class_getitem__(cls, annotation: Any) -> type[Self]:
        field_definition = FieldDefinition.from_annotation(annotation)

        if (field_definition.is_optional and len(field_definition.args) > 2) or (
            field_definition.is_union and not field_definition.is_optional
        ):
            raise InvalidAnnotationException(
                "Unions are currently not supported as type argument to DTO. Want this? Open an issue."
            )

        if field_definition.is_forward_ref:
            raise InvalidAnnotationException("Forward references are not supported as type argument to DTO")

        # if a configuration is not provided, and the type narrowing is a type var, we don't want to create a subclass
        config = get_dto_config_from_annotated_type(field_definition)

        if not config:
            if field_definition.is_type_var:
                return cls
            config = cls.config if hasattr(cls, "config") else DTOConfig()

        cls_dict: dict[str, Any] = {"config": config, "_type_backend_map": {}, "_handler_backend_map": {}}
        if not field_definition.is_type_var:
            cls_dict.update(model_type=field_definition.annotation)

        return type(f"{cls.__name__}[{annotation}]", (cls,), cls_dict)

    def builtins_to_data_type(self, builtins: Any) -> Any:
        """Coerce the unstructured data into the data type."""
        backend = self._get_backend("data", self.connection_context.handler_id)
        return backend.populate_data_from_builtins(builtins, self.connection_context)

    def bytes_to_data_type(self, raw: bytes) -> Any:
        """Return the data held by the DTO."""
        backend = self._get_backend("data", self.connection_context.handler_id)
        return backend.populate_data_from_raw(raw, self.connection_context)

    def data_to_encodable_type(self, data: T | Collection[T]) -> LitestarEncodableType:
        backend = self._get_backend("return", self.connection_context.handler_id)
        return backend.encode_data(data, self.connection_context)

    @classmethod
    @abstractmethod
    def generate_field_definitions(cls, model_type: type[Any]) -> Generator[DTOFieldDefinition, None, None]:
        """Generate ``FieldDefinition`` instances from ``model_type``.

        Yields:
            ``FieldDefinition`` instances.
        """

    @classmethod
    @abstractmethod
    def detect_nested_field(cls, field_definition: FieldDefinition) -> bool:
        """Return ``True`` if ``field_definition`` represents a nested model field.

        Args:
            field_definition: inspect type to determine if field represents a nested model.

        Returns:
            ``True`` if ``field_definition`` represents a nested model field.
        """

    @classmethod
    def on_registration(cls, handler_context: HandlerContext) -> None:
        """Called each time the DTO type is encountered during signature modelling.

        Args:
            handler_context: A :class:`HandlerContext <.HandlerContext>` instance. Provides information about the
                handler and application of the DTO.
        """
        field_definition = handler_context.field_definition
        model_type = resolve_model_type(field_definition)
        wrapper_attribute_name: str | None = None

        if not model_type.is_subclass_of(cls.model_type):
            resolved_generic_result = resolve_generic_wrapper_type(model_type, cls.model_type)
            if resolved_generic_result is not None:
                model_type, field_definition, wrapper_attribute_name = resolved_generic_result
            else:
                raise InvalidAnnotationException(
                    f"DTO narrowed with '{cls.model_type}', handler type is '{field_definition.annotation}'"
                )

        key = (handler_context.dto_for, field_definition, handler_context.request_encoding_type)
        backend = cls._type_backend_map.get(key)
        if backend is None:
            backend_type: type[DTOBackend] = DTOBackend

            backend_context = BackendContext(
                dto_config=cls.config,
                dto_for=handler_context.dto_for,
                field_definition=field_definition,
                field_definition_generator=cls.generate_field_definitions,
                is_nested_field_predicate=cls.detect_nested_field,
                model_type=model_type.annotation,
                wrapper_attribute_name=wrapper_attribute_name,
            )
            backend = cls._type_backend_map.setdefault(key, backend_type(backend_context))
        cls._handler_backend_map[(handler_context.dto_for, handler_context.handler_id)] = backend

    @classmethod
    def create_openapi_schema(
        cls, dto_for: ForType, handler_id: str, schema_creator: SchemaCreator
    ) -> Reference | Schema:
        """Create an OpenAPI request body.

        Returns:
            OpenAPI request body.
        """
        return cls._get_backend(dto_for, handler_id).create_openapi_schema(schema_creator)

    @classmethod
    def _get_backend(cls, dto_for: ForType, handler_id: str) -> DTOBackend:
        """Return the backend for the handler/dto_for combo."""
        return cls._handler_backend_map[(dto_for, handler_id)]



from litestar.utils import warn_deprecation


def __getattr__(attr_name: str) -> object:
    if "DataclassDTO" in attr_name:
        from litestar.dto import DataclassDTO

        warn_deprecation(
            deprecated_name="litestar.dto.factory.stdlib.dataclass.DataclassDTO",
            version="2.0b3",
            kind="import",
            removal_in="2.0",
            info="importing 'DataclassDTO' from 'litestar.dto.factory.stdlib.dataclass' is deprecated, please"
            "import it from 'litestar.dto.factory' instead",
        )

        globals()[attr_name] = DataclassDTO
        return DataclassDTO
    raise AttributeError(f"module {__name__!r} has no attribute {attr_name!r}")



from __future__ import annotations

import typing
from inspect import getmodule
from typing import TYPE_CHECKING, Collection, Mapping, TypeVar, Union, cast

from msgspec import UNSET, Struct, UnsetType, defstruct, field
from typing_extensions import get_origin, get_type_hints

from litestar.dto._types import (
    CollectionType,
    CompositeType,
    MappingType,
    NestedFieldInfo,
    SimpleType,
    TransferType,
    TupleType,
    UnionType,
)
from litestar.dto.config import DTOConfig
from litestar.dto.data_structures import DTOData
from litestar.dto.field import Mark
from litestar.dto.types import ForType, RenameStrategy
from litestar.enums import RequestEncodingType
from litestar.types import Empty
from litestar.types.builtin_types import NoneType
from litestar.types.composite_types import TypeEncodersMap
from litestar.types.protocols import InstantiableCollection
from litestar.typing import FieldDefinition
from litestar.utils import find_index

if TYPE_CHECKING:
    from typing import AbstractSet, Any, Iterable

    from litestar.dto._types import FieldDefinitionsType, TransferDTOFieldDefinition
    from litestar.dto.data_structures import DTOFieldDefinition

__all__ = (
    "RenameStrategies",
    "build_annotation_for_backend",
    "create_msgspec_field",
    "create_struct_for_field_definitions",
    "create_transfer_model_type_annotation",
    "get_dto_config_from_annotated_type",
    "get_model_type_hints",
    "resolve_generic_wrapper_type",
    "resolve_model_type",
    "should_exclude_field",
    "should_ignore_field",
    "should_mark_private",
    "transfer_data",
    "transfer_nested_union_type_data",
)

T = TypeVar("T")
StructT = TypeVar("StructT", bound=Struct)


def get_model_type_hints(model_type: type[Any], namespace: dict[str, Any] | None = None) -> dict[str, FieldDefinition]:
    """Retrieve type annotations for ``model_type``.

    Args:
        model_type: Any type-annotated class.
        namespace: Optional namespace to use for resolving type hints.

    Returns:
        Parsed type hints for ``model_type`` resolved within the scope of its module.
    """
    namespace = namespace or {}
    namespace.update(vars(typing))
    namespace.update(
        {
            "TypeEncodersMap": TypeEncodersMap,
            "ForType": ForType,
            "DTOConfig": DTOConfig,
            "RenameStrategy": RenameStrategy,
            "RequestEncodingType": RequestEncodingType,
        }
    )

    if model_module := getmodule(model_type):
        namespace.update(vars(model_module))

    return {
        k: FieldDefinition.from_kwarg(annotation=v, name=k)
        for k, v in get_type_hints(model_type, localns=namespace, include_extras=True).items()
    }


def get_dto_config_from_annotated_type(field_definition: FieldDefinition) -> DTOConfig | None:
    """Extract data type and config instances from ``Annotated`` annotation.

    Args:
        field_definition: A parsed type annotation that represents the annotation used to narrow the DTO type.

    Returns:
        The type and config object extracted from the annotation.
    """
    if configs := [item for item in field_definition.metadata if isinstance(item, DTOConfig)]:
        return configs[0]
    return None


def resolve_model_type(field_definition: FieldDefinition) -> FieldDefinition:
    """Resolve the data model type from a parsed type.

    Args:
        field_definition: A parsed type annotation that represents the annotation used to narrow the DTO type.

    Returns:
        A :class:`FieldDefinition <.typing.FieldDefinition>` that represents the data model type.
    """
    if field_definition.is_optional:
        return resolve_model_type(next(t for t in field_definition.inner_types if not t.is_subclass_of(NoneType)))

    if field_definition.is_subclass_of(DTOData):
        return resolve_model_type(field_definition.inner_types[0])

    if field_definition.is_collection:
        if field_definition.is_mapping:
            return resolve_model_type(field_definition.inner_types[1])

        if field_definition.is_tuple:
            if any(t is Ellipsis for t in field_definition.args):
                return resolve_model_type(field_definition.inner_types[0])
        elif field_definition.is_non_string_collection:
            return resolve_model_type(field_definition.inner_types[0])

    return field_definition


def resolve_generic_wrapper_type(
    field_definition: FieldDefinition, dto_specialized_type: type[Any]
) -> tuple[FieldDefinition, FieldDefinition, str] | None:
    """Handle where DTO supported data is wrapped in a generic container type.

    Args:
        field_definition: A parsed type annotation that represents the annotation used to narrow the DTO type.
        dto_specialized_type: The type used to specialize the DTO.

    Returns:
        The data model type.
    """
    if (origin := field_definition.origin) and (parameters := getattr(origin, "__parameters__", None)):
        param_index = find_index(
            field_definition.inner_types, lambda x: resolve_model_type(x).is_subclass_of(dto_specialized_type)
        )

        if param_index == -1:
            return None

        inner_type = field_definition.inner_types[param_index]
        model_type = resolve_model_type(inner_type)
        type_var = parameters[param_index]

        for attr, attr_type in get_model_type_hints(origin).items():
            if attr_type.annotation is type_var or any(t.annotation is type_var for t in attr_type.inner_types):
                if attr_type.is_non_string_collection:
                    # the inner type of the collection type is the type var, so we need to specialize the
                    # collection type with the DTO supported type.
                    specialized_annotation = attr_type.safe_generic_origin[model_type.annotation]
                    return model_type, FieldDefinition.from_annotation(specialized_annotation), attr
                return model_type, inner_type, attr

    return None


def build_annotation_for_backend(annotation: Any, model: type[T]) -> type[T] | type[Iterable[T]]:
    """A helper to re-build a generic outer type with new inner type.

    Args:
        annotation: The original annotation on the handler signature
        model: The data container type

    Returns:
        Annotation with new inner type if applicable.
    """
    origin = get_origin(annotation)
    if not origin:
        return model
    try:
        return origin[model]  # type:ignore[no-any-return]
    except TypeError:  # pragma: no cover
        return annotation.copy_with((model,))  # type:ignore[no-any-return]


def should_mark_private(field_definition: DTOFieldDefinition, underscore_fields_private: bool) -> bool:
    """Returns ``True`` where a field should be marked as private.

    Fields should be marked as private when:
    - the ``underscore_fields_private`` flag is set.
    - the field is not already marked.
    - the field name is prefixed with an underscore

    Args:
        field_definition: defined DTO field
        underscore_fields_private: whether fields prefixed with an underscore should be marked as private.
    """
    return bool(
        underscore_fields_private and field_definition.dto_field.mark is None and field_definition.name.startswith("_")
    )


def should_exclude_field(
    field_definition: DTOFieldDefinition, exclude: AbstractSet[str], include: AbstractSet[str], dto_for: ForType
) -> bool:
    """Returns ``True`` where a field should be excluded from data transfer.

    Args:
        field_definition: defined DTO field
        exclude: names of fields to exclude
        include: names of fields to exclude
        dto_for: indicates whether the DTO is for the request body or response.

    Returns:
        ``True`` if the field should not be included in any data transfer.
    """
    field_name = field_definition.name
    if field_name in exclude:
        return True
    if include and field_name not in include and not (any(f.startswith(f"{field_name}.") for f in include)):
        return True
    if field_definition.dto_field.mark is Mark.PRIVATE:
        return True
    if dto_for == "data" and field_definition.dto_field.mark is Mark.READ_ONLY:
        return True
    return dto_for == "return" and field_definition.dto_field.mark is Mark.WRITE_ONLY


def should_ignore_field(field_definition: DTOFieldDefinition, dto_for: ForType) -> bool:
    """Returns ``True`` where a field should be ignored.

    An ignored field is different to an excluded one in that we do not produce a
    ``TransferFieldDefinition`` for it at all.

    This allows ``AbstractDTOFactory`` concrete types to generate multiple field definitions
    for the same field name, each for a different transfer direction.

    One example of this is the :class:`sqlalchemy.ext.hybrid.hybrid_property` which, might have
    a different type accepted by its setter method, than is returned by its getter method.
    """
    return field_definition.dto_for is not None and field_definition.dto_for != dto_for


class RenameStrategies:
    """Useful renaming strategies than be used with :class:`DTOConfig`"""

    def __init__(self, renaming_strategy: RenameStrategy) -> None:
        self.renaming_strategy = renaming_strategy

    def __call__(self, field_name: str) -> str:
        if not isinstance(self.renaming_strategy, str):
            return self.renaming_strategy(field_name)

        return cast(str, getattr(self, self.renaming_strategy)(field_name))

    @staticmethod
    def upper(field_name: str) -> str:
        return field_name.upper()

    @staticmethod
    def lower(field_name: str) -> str:
        return field_name.lower()

    @staticmethod
    def camel(field_name: str) -> str:
        return RenameStrategies._camelize(field_name)

    @staticmethod
    def pascal(field_name: str) -> str:
        return RenameStrategies._camelize(field_name, capitalize_first_letter=True)

    @staticmethod
    def _camelize(string: str, capitalize_first_letter: bool = False) -> str:
        """Convert a string to camel case.

        Args:
            string (str): The string to convert
            capitalize_first_letter (bool): Default is False, a True value will convert to PascalCase
        Returns:
            str: The string converted to camel case or Pascal case
        """
        return "".join(
            word if index == 0 and not capitalize_first_letter else word.capitalize()
            for index, word in enumerate(string.split("_"))
        )


def transfer_data(
    destination_type: type[T],
    source_data: Any | Collection[Any],
    field_definitions: FieldDefinitionsType,
    dto_for: ForType,
    field_definition: FieldDefinition,
) -> T | InstantiableCollection[T]:
    """Create instance or iterable of instances of ``destination_type``.

    Args:
        destination_type: the model type received by the DTO on type narrowing.
        source_data: data that has been parsed and validated via the backend.
        field_definitions: model field definitions.
        dto_for: indicates whether the DTO is for the request body or response.
        field_definition: the parsed type that represents the handler annotation for which the DTO is being applied.

    Returns:
        Data parsed into ``destination_type``.
    """
    if field_definition.is_non_string_collection and not field_definition.is_mapping:
        origin = field_definition.instantiable_origin
        if not issubclass(origin, InstantiableCollection):  # pragma: no cover
            raise RuntimeError(
                f"Unexpected origin type '{field_definition.instantiable_origin}', expected collection type"
            )

        return origin(  # type:ignore[no-any-return]
            transfer_data(destination_type, item, field_definitions, dto_for, field_definition.inner_types[0])
            for item in source_data
        )
    return transfer_instance_data(destination_type, source_data, field_definitions, dto_for)


def transfer_instance_data(
    destination_type: type[T], source_instance: Any, field_definitions: FieldDefinitionsType, dto_for: ForType
) -> T:
    """Create instance of ``destination_type`` with data from ``source_instance``.

    Args:
        destination_type: the model type received by the DTO on type narrowing.
        source_instance: primitive data that has been parsed and validated via the backend.
        field_definitions: model field definitions.
        dto_for: indicates whether the DTO is for the request body or response.

    Returns:
        Data parsed into ``model_type``.
    """
    unstructured_data = {}
    source_is_mapping = isinstance(source_instance, Mapping)

    def has(source: Any, key: str) -> bool:
        return key in source if source_is_mapping else hasattr(source, key)

    def get(source: Any, key: str) -> Any:
        return source[key] if source_is_mapping else getattr(source, key)

    def filter_missing(value: Any) -> bool:
        return value is UNSET

    for field_definition in field_definitions:
        source_name = field_definition.serialization_name if dto_for == "data" else field_definition.name

        if should_skip_transfer(dto_for, field_definition, has(source_instance, source_name)):
            continue

        transfer_type = field_definition.transfer_type
        destination_name = field_definition.name if dto_for == "data" else field_definition.serialization_name
        source_value = get(source_instance, source_name)

        if field_definition.is_partial and dto_for == "data" and filter_missing(source_value):
            continue

        unstructured_data[destination_name] = transfer_type_data(
            source_value, transfer_type, dto_for, nested_as_dict=destination_type is dict
        )
    return destination_type(**unstructured_data)


def should_skip_transfer(
    dto_for: ForType,
    field_definition: TransferDTOFieldDefinition,
    source_has_value: bool,
) -> bool:
    """Returns ``True`` where a field should be excluded from data transfer.

    We should skip transfer when:
    - the field is excluded and the DTO is for the return data.
    - the DTO is for request data, and the field is not in the source instance.

    Args:
        dto_for: indicates whether the DTO is for the request body or response.
        field_definition: model field definition.
        source_has_value: indicates whether the source instance has a value for the field.
    """
    return (dto_for == "return" and field_definition.is_excluded) or (dto_for == "data" and not source_has_value)


def transfer_type_data(
    source_value: Any, transfer_type: TransferType, dto_for: ForType, nested_as_dict: bool = False
) -> Any:
    if isinstance(transfer_type, SimpleType) and transfer_type.nested_field_info:
        if nested_as_dict:
            dest_type = dict
        else:
            dest_type = (
                transfer_type.field_definition.annotation
                if dto_for == "data"
                else transfer_type.nested_field_info.model
            )

        return transfer_nested_simple_type_data(dest_type, transfer_type.nested_field_info, dto_for, source_value)
    if isinstance(transfer_type, UnionType) and transfer_type.has_nested:
        return transfer_nested_union_type_data(transfer_type, dto_for, source_value)
    if isinstance(transfer_type, CollectionType):
        if transfer_type.has_nested:
            return transfer_nested_collection_type_data(
                transfer_type.field_definition.instantiable_origin, transfer_type, dto_for, source_value
            )
        return transfer_type.field_definition.instantiable_origin(source_value)
    return source_value


def transfer_nested_collection_type_data(
    origin_type: type[Any], transfer_type: CollectionType, dto_for: ForType, source_value: Any
) -> Any:
    return origin_type(transfer_type_data(item, transfer_type.inner_type, dto_for) for item in source_value)


def transfer_nested_simple_type_data(
    destination_type: type[T],  # pyright: ignore
    nested_field_info: NestedFieldInfo,
    dto_for: ForType,
    source_value: Any,
) -> Any:
    return transfer_instance_data(destination_type, source_value, nested_field_info.field_definitions, dto_for)


def transfer_nested_union_type_data(transfer_type: UnionType, dto_for: ForType, source_value: Any) -> Any:
    for inner_type in transfer_type.inner_types:
        if isinstance(inner_type, CompositeType):
            raise RuntimeError("Composite inner types not (yet) supported for nested unions.")

        if inner_type.nested_field_info and isinstance(
            source_value,
            inner_type.nested_field_info.model if dto_for == "data" else inner_type.field_definition.annotation,
        ):
            return transfer_instance_data(
                inner_type.field_definition.annotation if dto_for == "data" else inner_type.nested_field_info.model,
                source_value,
                inner_type.nested_field_info.field_definitions,
                dto_for,
            )
    return source_value


def create_transfer_model_type_annotation(transfer_type: TransferType) -> Any:
    """Create a type annotation for a transfer model.

    Uses the parsed type that originates from the data model and the transfer model generated to represent a nested
    type to reconstruct the type annotation for the transfer model.
    """
    if isinstance(transfer_type, SimpleType):
        if transfer_type.nested_field_info:
            return transfer_type.nested_field_info.model
        return transfer_type.field_definition.annotation

    if isinstance(transfer_type, CollectionType):
        return create_transfer_model_collection_type(transfer_type)

    if isinstance(transfer_type, MappingType):
        return create_transfer_model_mapping_type(transfer_type)

    if isinstance(transfer_type, TupleType):
        return create_transfer_model_tuple_type(transfer_type)

    if isinstance(transfer_type, UnionType):
        return create_transfer_model_union_type(transfer_type)

    raise RuntimeError(f"Unexpected transfer type: {type(transfer_type)}")


def create_transfer_model_collection_type(transfer_type: CollectionType) -> Any:
    generic_collection_type = transfer_type.field_definition.safe_generic_origin
    inner_type = create_transfer_model_type_annotation(transfer_type.inner_type)
    if transfer_type.field_definition.origin is tuple:
        return generic_collection_type[inner_type, ...]
    return generic_collection_type[inner_type]


def create_transfer_model_tuple_type(transfer_type: TupleType) -> Any:
    inner_types = tuple(create_transfer_model_type_annotation(t) for t in transfer_type.inner_types)
    return transfer_type.field_definition.safe_generic_origin[inner_types]


def create_transfer_model_union_type(transfer_type: UnionType) -> Any:
    inner_types = tuple(create_transfer_model_type_annotation(t) for t in transfer_type.inner_types)
    return transfer_type.field_definition.safe_generic_origin[inner_types]


def create_transfer_model_mapping_type(transfer_type: MappingType) -> Any:
    key_type = create_transfer_model_type_annotation(transfer_type.key_type)
    value_type = create_transfer_model_type_annotation(transfer_type.value_type)
    return transfer_type.field_definition.safe_generic_origin[key_type, value_type]


def create_msgspec_field(field_definition: TransferDTOFieldDefinition) -> Any:
    kwargs: dict[str, Any] = {}
    if field_definition.is_partial:
        kwargs["default"] = UNSET

    elif field_definition.default is not Empty:
        kwargs["default"] = field_definition.default

    elif field_definition.default_factory is not None:
        kwargs["default_factory"] = field_definition.default_factory

    return field(**kwargs)


def create_struct_for_field_definitions(model_name: str, field_definitions: FieldDefinitionsType) -> type[Struct]:
    struct_fields: list[tuple[str, type] | tuple[str, type, type]] = []
    for field_def in field_definitions:
        if field_def.is_excluded:
            continue

        field_name = field_def.serialization_name or field_def.name

        field_type = create_transfer_model_type_annotation(field_def.transfer_type)
        if field_def.is_partial:
            field_type = Union[field_type, UnsetType]

        struct_fields.append((field_name, field_type, create_msgspec_field(field_def)))
    return defstruct(model_name, struct_fields, frozen=True, kw_only=True)



from typing import Any

from litestar.utils import warn_deprecation


def __getattr__(attr_name: str) -> object:
    if "DataclassDTO" in attr_name:
        from litestar.dto import DataclassDTO

        value: Any = DataclassDTO
    elif "AbstractDTOFactory" in attr_name:
        from litestar.dto.base_factory import AbstractDTOFactory

        value = AbstractDTOFactory
    elif "DTOData" in attr_name:
        from litestar.dto import DTOData

        value = DTOData
    elif "DTOConfig" in attr_name:
        from litestar.dto import DTOConfig

        value = DTOConfig

    elif "DTOField" in attr_name:
        from litestar.dto import DTOField

        value = DTOField

    elif "Mark" in attr_name:
        from litestar.dto import Mark

        value = Mark
    elif "dto_field" in attr_name:
        from litestar.dto import dto_field

        value = dto_field
    else:
        raise AttributeError(f"module {__name__!r} has no attribute {attr_name!r}")

    warn_deprecation(
        deprecated_name=f"litestar.dto.factory.{attr_name}",
        version="2.0b3",
        kind="import",
        removal_in="2.0",
        info=f"importing {attr_name} from 'litestar.dto.factory' is deprecated, please"
        f"import it from 'litestar.dto' instead",
    )

    globals()[attr_name] = value
    return value



from __future__ import annotations

import sys
import typing
from dataclasses import dataclass, replace
from inspect import Signature, getmembers, isclass, ismethod
from itertools import chain
from typing import Any

from typing_extensions import Self, get_type_hints

from litestar import connection, datastructures, types
from litestar.enums import RequestEncodingType
from litestar.params import BodyKwarg
from litestar.types import Empty
from litestar.typing import FieldDefinition

if typing.TYPE_CHECKING:
    from litestar.types import AnyCallable

_GLOBAL_NAMES = {
    namespace: export
    for namespace, export in chain(
        tuple(getmembers(types)), tuple(getmembers(connection)), tuple(getmembers(datastructures))
    )
    if namespace[0].isupper()
    and namespace in chain(types.__all__, connection.__all__, datastructures.__all__)  # pyright: ignore
}
"""A mapping of names used for handler signature forward-ref resolution.

This allows users to include these names within an `if TYPE_CHECKING:` block in their handler module.
"""

__all__ = (
    "get_fn_type_hints",
    "ParsedSignature",
    "infer_request_encoding_from_field_definition",
)


def get_fn_type_hints(fn: Any, namespace: dict[str, Any] | None = None) -> dict[str, Any]:
    """Resolve type hints for ``fn``.

    Args:
        fn: Callable that is being inspected
        namespace: Extra names for resolution of forward references.

    Returns:
        Mapping of names to types.
    """
    fn_to_inspect: Any = fn

    module_name = fn_to_inspect.__module__

    if isclass(fn_to_inspect):
        fn_to_inspect = fn_to_inspect.__init__

    # detect objects that are not functions and that have a `__call__` method
    if callable(fn_to_inspect) and ismethod(fn_to_inspect.__call__):
        fn_to_inspect = fn_to_inspect.__call__

    # inspect the underlying function for methods
    if hasattr(fn_to_inspect, "__func__"):
        fn_to_inspect = fn_to_inspect.__func__

    # Order important. If a litestar name has been overridden in the function module, we want
    # to use that instead of the litestar one.
    namespace = {
        **_GLOBAL_NAMES,
        **vars(typing),
        **vars(sys.modules[module_name]),
        **(namespace or {}),
    }
    return get_type_hints(fn_to_inspect, globalns=namespace, include_extras=True)


@dataclass(frozen=True)
class ParsedSignature:
    """Parsed signature.

    This object is the primary source of handler/dependency signature information.

    The only post-processing that occurs is the conversion of any forward referenced type annotations.
    """

    __slots__ = ("parameters", "return_type", "original_signature")

    parameters: dict[str, FieldDefinition]
    """A mapping of parameter names to ParsedSignatureParameter instances."""
    return_type: FieldDefinition
    """The return annotation of the callable."""
    original_signature: Signature
    """The raw signature as returned by :func:`inspect.signature`"""

    @classmethod
    def from_fn(cls, fn: AnyCallable, signature_namespace: dict[str, Any]) -> Self:
        """Parse a function signature.

        Args:
            fn: Any callable.
            signature_namespace: mapping of names to types for forward reference resolution

        Returns:
            ParsedSignature
        """
        signature = Signature.from_callable(fn)
        fn_type_hints = get_fn_type_hints(fn, namespace=signature_namespace)

        parameters = tuple(
            FieldDefinition.from_parameter(parameter=parameter, fn_type_hints=fn_type_hints)
            for name, parameter in signature.parameters.items()
            if name not in ("self", "cls")
        )

        return_type = FieldDefinition.from_annotation(fn_type_hints.get("return", Any))

        return cls(
            parameters={p.name: p for p in parameters},
            return_type=return_type if "return" in fn_type_hints else replace(return_type, annotation=Empty),
            original_signature=signature,
        )

    @classmethod
    def from_signature(cls, signature: Signature, signature_namespace: dict[str, Any]) -> Self:
        """Parse an :class:`inspect.Signature` instance.

        Python's `get_type_hints()` function does not support parsing signatures directly, so we need to create a dummy
        function to pass to it. Maybe there's a better way to do this, but this does work.

        Args:
            signature: An :class:`inspect.Signature` instance.
            signature_namespace: mapping of names to types for forward reference resolution

        Returns:
            ParsedSignature
        """

        def fn() -> None:
            ...

        fn.__signature__ = signature  # type:ignore[attr-defined]
        fn.__annotations__ = {p.name: p.annotation for p in signature.parameters.values()}
        return cls.from_fn(fn, signature_namespace)


def infer_request_encoding_from_field_definition(field_definition: FieldDefinition) -> RequestEncodingType | str:
    """Infer the request encoding type from a parsed type.

    Args:
        field_definition: The parsed parameter to infer the request encoding type from.

    Returns:
        The inferred request encoding type.
    """
    if field_definition.kwarg_definition and isinstance(field_definition.kwarg_definition, BodyKwarg):
        return field_definition.kwarg_definition.media_type
    if isinstance(field_definition.default, BodyKwarg):
        return field_definition.default.media_type
    return RequestEncodingType.JSON



from .base import Response
from .file import File
from .redirect import Redirect
from .streaming import Stream
from .template import Template

__all__ = ("Response", "Redirect", "Stream", "Template", "File")



from __future__ import annotations

from contextlib import contextmanager
from typing import Any

from sqlalchemy.exc import IntegrityError, SQLAlchemyError

from litestar.contrib.repository import ConflictError, RepositoryError


@contextmanager
def wrap_sqlalchemy_exception() -> Any:
    """Do something within context to raise a `RepositoryError` chained
    from an original `SQLAlchemyError`.

        >>> try:
        ...     with wrap_sqlalchemy_exception():
        ...         raise SQLAlchemyError("Original Exception")
        ... except RepositoryError as exc:
        ...     print(f"caught repository exception from {type(exc.__context__)}")
        ...
        caught repository exception from <class 'sqlalchemy.exc.SQLAlchemyError'>
    """
    try:
        yield
    except IntegrityError as exc:
        raise ConflictError from exc
    except SQLAlchemyError as exc:
        raise RepositoryError(f"An exception occurred: {exc}") from exc
    except AttributeError as exc:
        raise RepositoryError from exc
